/**
 * @file
 *
 * @ingroup rtems_bsd_rtems
 *
 * @brief TODO.
 *
 * File origin from FreeBSD 'lib/libc/gen/sysctlnametomib.c'.
 */

/*
 * Copyright 2001 The FreeBSD Project. All Rights Reserved.
 * Copyright 2020 Chris Johns
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE FREEBSD PROJECT BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <machine/rtems-bsd-kernel-space.h>
#include <machine/rtems-bsd-syscall-api.h>

#include <sys/file.h>
#include <sys/filedesc.h>
#include <sys/proc.h>
#include <sys/syscallsubr.h>
#include <sys/sysproto.h>
#include <sys/vnode.h>

#include <errno.h>
#include <sys/dirent.h>

#include <rtems/libio.h>
#include <rtems/seterr.h>

#include <machine/rtems-bsd-libio.h>
#include <machine/rtems-bsd-vfs.h>

static int rtems_bsd_sysgen_open_error(rtems_libio_t *iop,
                                       const char *path,
                                       int oflag,
                                       mode_t mode);
static int rtems_bsd_sysgen_open(rtems_libio_t *iop,
                                 const char    *path,
                                 int            oflag,
                                 mode_t         mode);
static int rtems_bsd_sysgen_close(rtems_libio_t *iop);
static ssize_t rtems_bsd_sysgen_read(rtems_libio_t *iop,
                                     void *buffer,
                                     size_t count);
static ssize_t rtems_bsd_sysgen_readv(rtems_libio_t      *iop,
                                      const struct iovec *iov,
                                      int                 iovcnt,
                                      ssize_t             total);
static ssize_t rtems_bsd_sysgen_write(rtems_libio_t *iop,
                                      const void *buffer,
                                      size_t count);
static ssize_t rtems_bsd_sysgen_writev(rtems_libio_t      *iop,
                                       const struct iovec *iov,
                                       int                 iovcnt,
                                       ssize_t             total);
static int rtems_bsd_sysgen_ioctl(rtems_libio_t *iop,
                                  ioctl_command_t request,
                                  void *buffer);
static off_t rtems_bsd_sysgen_lseek(rtems_libio_t *iop,
                                    off_t offset,
                                    int whence);
static int rtems_bsd_sysgen_vnstat(const rtems_filesystem_location_info_t *loc,
                                   struct stat *buf);
static int rtems_bsd_sysgen_fstat(const rtems_filesystem_location_info_t *loc,
                                  struct stat *buf);
static int rtems_bsd_sysgen_ftruncate(rtems_libio_t *iop, off_t length);
static int rtems_bsd_sysgen_fsync(rtems_libio_t *iop);
static int rtems_bsd_sysgen_fdatasync(rtems_libio_t *iop);
static int rtems_bsd_sysgen_fcntl(rtems_libio_t *iop, int cmd);
static int rtems_bsd_sysgen_poll(rtems_libio_t *iop, int events);
static int rtems_bsd_sysgen_kqfilter(rtems_libio_t *iop, struct knote *kn);

static const rtems_filesystem_file_handlers_r api_sysgen_dirops =
{
  .open_h      = rtems_bsd_sysgen_open,
  .close_h     = rtems_bsd_sysgen_close,
  .read_h      = rtems_bsd_sysgen_read,
  .write_h     = rtems_filesystem_default_write,
  .ioctl_h     = rtems_filesystem_default_ioctl,
  .lseek_h     = rtems_filesystem_default_lseek_directory,
  .fstat_h     = rtems_bsd_sysgen_vnstat,
  .ftruncate_h = rtems_filesystem_default_ftruncate_directory,
  .fsync_h     = rtems_filesystem_default_fsync_or_fdatasync,
  .fdatasync_h = rtems_bsd_sysgen_fdatasync,
  .fcntl_h     = rtems_filesystem_default_fcntl,
  .kqfilter_h  = rtems_filesystem_default_kqfilter,
  .mmap_h      = rtems_filesystem_default_mmap,
  .poll_h      = rtems_filesystem_default_poll,
  .readv_h     = rtems_filesystem_default_readv,
  .writev_h    = rtems_filesystem_default_writev
};

const rtems_filesystem_file_handlers_r* rtems_bsd_sysgen_dirops = &api_sysgen_dirops;

static const rtems_filesystem_file_handlers_r api_sysgen_fileops =
{
  .open_h      = rtems_bsd_sysgen_open,
  .close_h     = rtems_bsd_sysgen_close,
  .read_h      = rtems_bsd_sysgen_read,
  .write_h     = rtems_bsd_sysgen_write,
  .ioctl_h     = rtems_bsd_sysgen_ioctl,
  .lseek_h     = rtems_bsd_sysgen_lseek,
  .fstat_h     = rtems_bsd_sysgen_vnstat,
  .ftruncate_h = rtems_bsd_sysgen_ftruncate,
  .fsync_h     = rtems_filesystem_default_fsync_or_fdatasync,
  .fdatasync_h = rtems_bsd_sysgen_fdatasync,
  .fcntl_h     = rtems_bsd_sysgen_fcntl,
  .kqfilter_h  = rtems_bsd_sysgen_kqfilter,
  .mmap_h      = rtems_filesystem_default_mmap,
  .poll_h      = rtems_bsd_sysgen_poll,
  .readv_h     = rtems_bsd_sysgen_readv,
  .writev_h    = rtems_bsd_sysgen_writev
};

const rtems_filesystem_file_handlers_r* rtems_bsd_sysgen_fileops = &api_sysgen_fileops;

static const rtems_filesystem_file_handlers_r api_sysgen_nodeops = {
  .open_h      = rtems_bsd_sysgen_open_error,
  .close_h     = rtems_bsd_sysgen_close,
  .read_h      = rtems_bsd_sysgen_read,
  .write_h     = rtems_bsd_sysgen_write,
  .ioctl_h     = rtems_bsd_sysgen_ioctl,
  .lseek_h     = rtems_filesystem_default_lseek,
  .fstat_h     = rtems_bsd_sysgen_fstat,
  .ftruncate_h = rtems_filesystem_default_ftruncate,
  .fsync_h     = rtems_filesystem_default_fsync_or_fdatasync,
  .fdatasync_h = rtems_filesystem_default_fsync_or_fdatasync,
  .fcntl_h     = rtems_bsd_sysgen_fcntl,
  .poll_h      = rtems_bsd_sysgen_poll,
  .kqfilter_h  = rtems_bsd_sysgen_kqfilter,
  .readv_h     = rtems_bsd_sysgen_readv,
  .writev_h    = rtems_bsd_sysgen_writev,
  .mmap_h      = rtems_filesystem_default_mmap
};

const rtems_filesystem_file_handlers_r* rtems_bsd_sysgen_nodeops = &api_sysgen_nodeops;

int
accept(int socket, struct sockaddr *__restrict address,
       socklen_t *__restrict address_len)
{
  struct thread *td;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: accept: %d\n", socket);
  td = rtems_bsd_get_curthread_or_null();
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct accept_args ua = {
        .s = ffd,
        .name = address,
        .anamelen = address_len
      };
      error = sys_accept(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
      if (error == 0) {
        rtems_libio_t *iop = rtems_bsd_libio_iop_allocate();
        int afd = td->td_retval[0];
        if (iop != NULL) {
          error = rtems_bsd_libio_iop_set_bsd_fd(td, afd,
                                                 iop, rtems_bsd_sysgen_nodeops);
          if (error == 0) {
            if (RTEMS_BSD_SYSCALL_TRACE)
              printf("bsd: sys: accept: %d (%d) => %d -> %d\n",
                     socket, ffd,
                     rtems_libio_iop_to_descriptor(iop),
                     rtems_bsd_libio_iop_to_descriptor(iop));
            return rtems_libio_iop_to_descriptor(iop);
          }
          rtems_bsd_libio_iop_free(iop);
        } else {
          error = ENFILE;
        }
        kern_close(td, afd);
      }
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
bind(int socket, const struct sockaddr *address, socklen_t address_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: bind: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd != -1) {
      struct bind_args ua = {
        .s = ffd,
        .name = address,
        .namelen = address_len
      };
      error = sys_bind(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
connect(int socket, const struct sockaddr *address, socklen_t address_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: connect: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd != -1) {
      struct connect_args ua = {
        .s = ffd,
        .name = address,
        .namelen = address_len
      };
      error = sys_connect(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
getpeername(int socket, struct sockaddr *__restrict address,
	    socklen_t *__restrict address_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: getpeername: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd != -1) {
      struct getpeername_args ua = {
        .fdes = ffd,
        .asa = address,
        .alen = address_len
      };
      error = sys_getpeername(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
getsockname(int socket, struct sockaddr *__restrict address,
	    socklen_t *__restrict address_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: getsockname: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct getsockname_args ua = {
        .fdes = ffd,
        .asa = address,
        .alen = address_len
      };
      error = sys_getsockname(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
getsockopt(int socket, int level, int option_name, void *__restrict
	   option_value, socklen_t *__restrict option_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: getsockopt: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct getsockopt_args ua = {
        .s = ffd,
        .level = level,
        .name = option_name,
        .val = (caddr_t) option_value,
        .avalsize = option_len
      };
      error = sys_getsockopt(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
kqueue(void)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: kqueue:\n");
  if (td != NULL) {
    rtems_libio_t *iop = rtems_bsd_libio_iop_allocate();
    if (iop != NULL) {
      struct kqueue_args ua = {
      };
      error = sys_kqueue(td, &ua);
      if (error == 0) {
        error = rtems_bsd_libio_iop_set_bsd_fd(td, td->td_retval[0],
                                               iop, rtems_bsd_sysgen_nodeops);
        if (error == 0) {
          if (RTEMS_BSD_SYSCALL_TRACE)
            printf("bsd: sys: kqueue: %d -> %d\n",
                   rtems_libio_iop_to_descriptor(iop),
                   rtems_bsd_libio_iop_to_descriptor(iop));
          return rtems_libio_iop_to_descriptor(iop);
        }
        kern_close(td, rtems_libio_iop_to_descriptor(iop));
      }
      rtems_bsd_libio_iop_free(iop);
    }
  } else {
    error = ENOMEM;
  }

  return rtems_bsd_error_to_status_and_errno(error);
}

__weak_reference(kevent, _kevent);

int
kevent(int kq, const struct kevent *changelist, int nchanges,
       struct kevent *eventlist, int nevents,
       const struct timespec *timeout)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: kevent: %d\n", kq);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(kq, NULL);
    if (ffd >= 0) {
      struct kevent_args ua = {
        .fd = ffd,
        .changelist = changelist,
        .nchanges = nchanges,
        .eventlist = eventlist,
        .nevents = nevents,
        .timeout = timeout
      };
      error = sys_kevent(td, &ua);
      rtems_bsd_libio_iop_drop(kq);
      if (error == 0) {
        return td->td_retval[0];
      }
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
listen(int socket, int backlog)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: listen: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct listen_args ua = {
        .s = ffd,
        .backlog = backlog
      };
      error = sys_listen(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
pipe(int fildes[2])
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: pipe: %d\n", socket);
  if (td != NULL) {
    rtems_libio_t *iop[2];
    iop[0] = rtems_bsd_libio_iop_allocate();
    if (iop[0] != NULL) {
      iop[1] = rtems_bsd_libio_iop_allocate();
      if (iop[1] != NULL) {
        error = kern_pipe(td, fildes, 0, NULL, NULL);
        if (error == 0) {
          error = rtems_bsd_libio_iop_set_bsd_fd(td, fildes[0],
                                                 iop[0], rtems_bsd_sysgen_nodeops);
          if (error == 0) {
            error = rtems_bsd_libio_iop_set_bsd_fd(td, fildes[1],
                                                 iop[1], rtems_bsd_sysgen_nodeops);
            if (error == 0) {
              fildes[0] = rtems_libio_iop_to_descriptor(iop[0]);
              fildes[1] = rtems_libio_iop_to_descriptor(iop[1]);
              if (RTEMS_BSD_SYSCALL_TRACE)
                printf("bsd: sys: pipe: %d -> %d, %d -> %d\n",
                       fildes[0], rtems_bsd_libio_iop_to_descriptor(iop[0]),
                       fildes[1], rtems_bsd_libio_iop_to_descriptor(iop[1]));
              return 0;
            }
          }
        }
        kern_close(td, rtems_bsd_libio_iop_to_descriptor(iop[0]));
        kern_close(td, rtems_bsd_libio_iop_to_descriptor(iop[1]));
        rtems_bsd_libio_iop_free(iop[0]);
        rtems_bsd_libio_iop_free(iop[1]);
      } else {
        rtems_bsd_libio_iop_free(iop[0]);
        error =ENFILE;
      }
    } else {
      error =ENFILE;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
poll(struct pollfd fds[], nfds_t nfds, int timeout)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: poll: %d\n", nfds);
  if (td != NULL) {
    /*
     * Pass libio descriptors through as libio and bsd descriptors can
     * be in the list at the same time.
     */
    struct poll_args ua = {
      .fds = &fds[0],
      .nfds = nfds,
      .timeout = timeout
    };
    error = sys_poll(td, &ua);
    if (error == 0) {
      return td->td_retval[0];
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
pselect(int nfds, fd_set *readfds, fd_set *writefds, fd_set *errorfds,
	const struct timespec *timeout, const sigset_t *set)
{
  struct thread *td;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: pselect: %d\n", nfds);
  if (set != NULL) {
    return rtems_bsd_error_to_status_and_errno(ENOSYS);
  }
  td = rtems_bsd_get_curthread_or_null();
  if (td != NULL) {
    struct timeval tv;
    struct timeval *tvp;
    if (timeout != NULL) {
      TIMESPEC_TO_TIMEVAL(&tv, timeout);
      tvp = &tv;
    } else {
      tvp = NULL;
    }
    /*
     * Pass libio descriptors through as libio and bsd descriptors can
     * be in the list at the same time.
     */
    error = kern_select(td, nfds, readfds, writefds, errorfds,
			tvp, NFDBITS);
    if (error == 0) {
      return td->td_retval[0];
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

ssize_t
recvfrom(int socket, void *__restrict buffer, size_t length, int flags,
	 struct sockaddr *__restrict address, socklen_t *__restrict address_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: recvfrom: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct recvfrom_args ua = {
        .s = ffd,
        .buf = buffer,
        .len = length,
        .flags = flags,
        .from = address,
        .fromlenaddr = address_len
      };
      error = sys_recvfrom(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
      if (error == 0) {
        return td->td_retval[0];
      }
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

ssize_t
recvmsg(int socket, struct msghdr *message, int flags)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: recvmsg: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct recvmsg_args ua = {
        .s = ffd,
        .msg = message,
        .flags = flags
      };
      error = sys_recvmsg(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
      if (error == 0) {
        return td->td_retval[0];
      }
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *errorfds,
       struct timeval *timeout)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: select: %d\n", nfds);
  if (td != NULL) {
    /*
     * Pass libio descriptors through as libio and bsd descriptors can
     * be in the list at the same time.
     */
    error = kern_select(td, nfds, readfds, writefds, errorfds,
			timeout, NFDBITS);
    if (error == 0) {
      return td->td_retval[0];
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

ssize_t
sendto(int socket, const void *message, size_t length, int flags,
       const struct sockaddr *dest_addr, socklen_t dest_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: sendto: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct sendto_args ua = {
        .s = ffd,
        .buf = (caddr_t) message,
        .len = length,
        .flags = flags,
        .to = dest_addr,
        .tolen = dest_len
      };
      error = sys_sendto(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
      if (error == 0) {
        return td->td_retval[0];
      }
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

ssize_t
sendmsg(int socket, const struct msghdr *message, int flags)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: sendmsg: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct sendmsg_args ua = {
        .s = ffd,
        .msg = message,
        .flags = flags
      };
      error = sys_sendmsg(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
setfib(int fibnum)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (td != NULL) {
    struct setfib_args ua = {
      .fibnum = fibnum
    };
    error = sys_setfib(td, &ua);
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
setsockopt(int socket, int level, int option_name, const void *option_value,
	   socklen_t option_len)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: setsockopt: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      struct setsockopt_args ua = {
        .s = ffd,
        .level = level,
        .name = option_name,
        .val = __DECONST(void *, option_value),
        .valsize = option_len
      };
      error = sys_setsockopt(td, &ua);
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
shutdown(int socket, int how)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: shutdown: %d\n", socket);
  if (td != NULL) {
    int ffd = rtems_bsd_libio_iop_hold(socket, NULL);
    if (ffd >= 0) {
      if (rtems_bsd_is_libbsd_descriptor(rtems_libio_iop(socket))) {
        struct shutdown_args ua = {
          .s = ffd,
          .how = how
        };
        error = sys_shutdown(td, &ua);
      } else {
        error = ENOTSOCK;
      }
      rtems_bsd_libio_iop_drop(socket);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
socket(int domain, int type, int protocol)
{
  struct thread *td;
  int error;
  td = rtems_bsd_get_curthread_or_null();
  if (td != NULL) {
    rtems_libio_t *iop = rtems_bsd_libio_iop_allocate();
    if (iop != NULL) {
      struct socket_args ua = {
        .domain = domain,
        .type = type,
        .protocol = protocol
      };
      error = sys_socket(td, &ua);
      if (error == 0) {
        error = rtems_bsd_libio_iop_set_bsd_fd(td, td->td_retval[0],
                                               iop, rtems_bsd_sysgen_nodeops);
        if (error == 0) {
          if (RTEMS_BSD_SYSCALL_TRACE)
            printf("bsd: sys: socket: %d -> %d\n",
                   rtems_libio_iop_to_descriptor(iop),
                   rtems_bsd_libio_iop_to_descriptor(iop));
          return rtems_libio_iop_to_descriptor(iop);
        }
        kern_close(td, td->td_retval[0]);
      }
      rtems_bsd_libio_iop_free(iop);
    } else {
      error = ENFILE;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
socketpair(int domain, int type, int protocol, int *socket_vector)
{
  struct thread *td;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: socketpair:\n");
  td = rtems_bsd_get_curthread_or_null();
  if (td != NULL) {
    rtems_libio_t *iop[2];
    iop[0] = rtems_bsd_libio_iop_allocate();
    if (iop[0] != NULL) {
      iop[1] = rtems_bsd_libio_iop_allocate();
      if (iop[1] != NULL) {
        struct socketpair_args ua = {
          .domain = domain,
          .type = type,
          .protocol = protocol,
          .rsv = socket_vector
        };
        error = sys_socketpair(td, &ua);
        if (error == 0) {
          error = rtems_bsd_libio_iop_set_bsd_fd(td,
                                                 socket_vector[0],
                                                 iop[0],
                                                 rtems_bsd_sysgen_nodeops);
          if (error == 0) {
            error = rtems_bsd_libio_iop_set_bsd_fd(td,
                                                   socket_vector[1],
                                                   iop[1],
                                                   rtems_bsd_sysgen_nodeops);
            if (error == 0) {
              socket_vector[0] = rtems_libio_iop_to_descriptor(iop[0]);
              socket_vector[1] = rtems_libio_iop_to_descriptor(iop[1]);
              printf("bsd: sys: socketpair: %d -> %d, %d -> %d\n",
                     socket_vector[0],
                     rtems_bsd_libio_iop_to_descriptor(iop[0]),
                     socket_vector[1],
                     rtems_bsd_libio_iop_to_descriptor(iop[1]));
              return 0;
            }
          }
        }
        kern_close(td, rtems_bsd_libio_iop_to_descriptor(iop[0]));
        kern_close(td, rtems_bsd_libio_iop_to_descriptor(iop[1]));
        rtems_bsd_libio_iop_free(iop[0]);
        rtems_bsd_libio_iop_free(iop[1]);
      }
      else {
        rtems_bsd_libio_iop_free(iop[0]);
        error = ENFILE;
      }
    } else {
      error = ENFILE;
    }
  } else {
    error = ENOMEM;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
sysctl(const int *name, u_int namelen, 	void *oldp, size_t *oldlenp,
       const void *newp, size_t newlen)
{
  int error = EINVAL;
  if (namelen <= CTL_MAXNAME) {
    int namedup [CTL_MAXNAME];
    memcpy(namedup, name, namelen * sizeof(*name));
    error = kernel_sysctl(NULL,
                          namedup,
                          namelen,
                          oldp,
                          oldlenp,
                          RTEMS_DECONST(void *, newp),
                          newlen,
                          oldlenp,
                          0);
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

/*
 * File origin from FreeBSD 'lib/libc/gen/sysctlbyname.c'.
 *
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.org> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 */
int
sysctlbyname(const char *name, void *oldp, size_t *oldlenp,
	     const void *newp, size_t newlen)
{
  int real_oid[CTL_MAXNAME+2];
  int error;
  size_t oidlen;
  oidlen = sizeof(real_oid) / sizeof(int);
  error = sysctlnametomib(name, real_oid, &oidlen);
  if (error < 0)
    return (error);
  error = sysctl(real_oid, oidlen, oldp, oldlenp, newp, newlen);
  return rtems_bsd_error_to_status_and_errno(error);
}

/*
 * File origin from FreeBSD 'lib/libc/gen/sysctlnametomib.c'.
 *
 * This function uses a presently undocumented interface to the kernel
 * to walk the tree and get the type so it can print the value.
 * This interface is under work and consideration, and should probably
 * be killed with a big axe by the first person who can find the time.
 * (be aware though, that the proper interface isn't as obvious as it
 * may seem, there are various conflicting requirements.
 */
int
sysctlnametomib(const char *name, int *mibp, size_t *sizep)
{
  int oid[2];
  int error;
  oid[0] = 0;
  oid[1] = 3;
  *sizep *= sizeof(int);
  error = sysctl(oid, 2, mibp, sizep, name, strlen(name));
  *sizep /= sizeof(int);
  return (error);
}

static int
rtems_bsd_sysgen_open_error(rtems_libio_t *iop,
                            const char *path,
                            int oflag,
                            mode_t mode)
{
  return rtems_bsd_error_to_status_and_errno(ENXIO);
}

int
rtems_bsd_sysgen_open(rtems_libio_t *iop,
                      const char    *path,
                      int            oflag,
                      mode_t         mode)
{
  struct thread *td = rtems_bsd_get_curthread_or_null();
  struct filedesc *fdp;
  struct file *fp;
  const char* opath;
  const bool creat = (oflag & O_CREAT) == O_CREAT;
  int pathlen;
  int fd;
  int error;

  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: open: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }

  fdp = td->td_proc->p_fd;
  pathlen = strlen(path);;

  /*
   * There is no easy or clean means to open a vnode and follow the
   * POSIX open semantics. You can open a vnode but the extra
   * functionality such as create and truncate are not part of the
   * basic vnode open. All the calls that provide that functionality
   * take a path as the argument. As a result find the last token in
   * the path and use the parent directory vnode to position ourselves
   * in the parent directory. The pathloc vnode points to the '.' or
   * '..'  directory.
   */
  if (rtems_filesystem_is_current_directory(path, pathlen) ||
      rtems_filesystem_is_parent_directory(path, pathlen)) {
    if (((oflag + 1) & _FWRITE) == _FWRITE) {
        if (RTEMS_BSD_SYSCALL_TRACE)
          printf("bsd: sys: open: write to .  or ..\n");
        return rtems_bsd_error_to_status_and_errno(EPERM);
    }
    opath = ".";
    fdp->fd_cdir = rtems_bsd_libio_loc_to_vnode(&iop->pathinfo);
  } else {
    opath = path + pathlen;
    while (opath != path && !rtems_filesystem_is_delimiter(opath[-1])) {
      opath--;
    }
    fdp->fd_cdir = rtems_bsd_libio_loc_to_vnode_dir(&iop->pathinfo);
  }

  fdp->fd_rdir = fdp->fd_cdir;

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: open: path=%s opath=%s vn=%p cwd=%p" \
           " flags=%08x mode=%08x\n",
           path, opath,
           creat ? NULL : rtems_bsd_libio_loc_to_vnode(&iop->pathinfo),
           fdp->fd_cdir,
           oflag, mode);

  VREF(fdp->fd_cdir);

  error = kern_openat(td, AT_FDCWD,
                      RTEMS_DECONST(char*, opath),
                      UIO_USERSPACE, oflag, mode);

  vrele(fdp->fd_cdir);

  if (error != 0) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: open: error = (%d) %s\n", error, strerror(error));
    return rtems_bsd_error_to_status_and_errno(error);
  }

  fd = td->td_retval[0];

  FILEDESC_XLOCK(fdp);
  if (fd < fdp->fd_nfiles) {
    struct vnode* vn;
    fp = fget_locked(fdp, fd);
    if (fp != NULL) {
      vn = fp->f_vnode;
    } else {
      vn = NULL;
    }
    rtems_bsd_libio_loc_set_vnode(&iop->pathinfo, vn);
  }
  FILEDESC_XUNLOCK(fdp);

  rtems_bsd_libio_iop_set_bsd_fd(td, fd,
                                 iop, rtems_bsd_sysgen_fileops);


  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: open: fd = %d vn=%p\n",
           fd, rtems_bsd_libio_loc_to_vnode(&iop->pathinfo));

  return 0;
}

int
rtems_bsd_sysgen_close(rtems_libio_t *iop)
{
  struct thread *td = curthread;
  int error;
  int ffd = rtems_bsd_libio_iop_to_descriptor(iop);
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: close: %d -> %d\n",
           rtems_libio_iop_to_descriptor(iop), ffd);
  if (td != NULL) {
    if (ffd >= 0) {
      rtems_libio_iop_hold(iop);
      error = kern_close(td, ffd);
    } else {
      error = EBADF;
    }
  } else {
    error = ENOMEM;
  }
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: close: %d: %d: %s\n",
           rtems_libio_iop_to_descriptor(iop), error, strerror(error));
  return rtems_bsd_error_to_status_and_errno(error);
}

ssize_t
rtems_bsd_sysgen_read(rtems_libio_t *iop, void *buffer, size_t count)
{
  struct thread *td = curthread;
  struct vnode *vp = rtems_bsd_libio_iop_to_vnode(iop);
  int fd = rtems_bsd_libio_iop_to_descriptor(iop);
  int error;
  ssize_t size = 0;

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: read: %d -> %d: vn=%p len=%d\n",
           rtems_libio_iop_to_descriptor(iop), fd, vp, count);

  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: read: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }

  if (count > IOSIZE_MAX)
    return rtems_bsd_error_to_status_and_errno(EINVAL);

  if (vp != NULL && vp->v_type == VDIR) {
    off_t offset;
    error = kern_getdirentries(td, fd, buffer, count,
                               &iop->offset, NULL, UIO_USERSPACE);
    size = td->td_retval[0];
  } else {
    struct iovec aiov = {
      .iov_base = buffer,
      .iov_len = count
    };
    struct uio auio = {
      .uio_iov = &aiov,
      .uio_iovcnt = 1,
      .uio_offset = iop->offset,
      .uio_resid = count,
      .uio_segflg = UIO_USERSPACE,
      .uio_rw = UIO_READ,
      .uio_td = td
    };
    error = kern_readv(td, rtems_bsd_libio_iop_to_descriptor(iop), &auio);
    if (error == 0)
      size = td->td_retval[0];
  }

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: read: %d: %d: %s size=%i\n",
           rtems_libio_iop_to_descriptor(iop), error, strerror(error), size);
  if (error != 0)
    return rtems_bsd_error_to_status_and_errno(error);

  return size;
}

ssize_t
rtems_bsd_sysgen_readv(rtems_libio_t      *iop,
                       const struct iovec *iov,
                       int                 iovcnt,
                       ssize_t             total)
{
  struct thread *td = curthread;
  struct uio auio;
  int error;

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: readv: %d len=%d\n", rtems_libio_iop_to_descriptor(iop), total);

  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: readv: read: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }

  if (total > IOSIZE_MAX)
    return rtems_bsd_error_to_status_and_errno(EINVAL);

  auio.uio_iov = RTEMS_DECONST(struct iovec*, iov);
  auio.uio_iovcnt = iovcnt;
  auio.uio_resid = total;
  auio.uio_segflg = UIO_USERSPACE;

  error = kern_readv(td, rtems_bsd_libio_iop_to_descriptor(iop),
		     &auio);

  if (error != 0)
    return rtems_bsd_error_to_status_and_errno(error);

  return td->td_retval[0];
}

ssize_t
rtems_bsd_sysgen_write(rtems_libio_t *iop, const void *buffer, size_t count)
{
  struct thread *td = curthread;
  struct uio auio;
  struct iovec aiov;
  int error;

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: write: %d len=%d\n", rtems_libio_iop_to_descriptor(iop), count);

  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: write: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }

  if (count > IOSIZE_MAX)
    return (EINVAL);

  aiov.iov_base = RTEMS_DECONST(void*, buffer);
  aiov.iov_len = count;
  auio.uio_iov = &aiov;
  auio.uio_iovcnt = 1;
  auio.uio_resid = count;
  auio.uio_segflg = UIO_USERSPACE;

  error = kern_writev(td, rtems_bsd_libio_iop_to_descriptor(iop),
		      &auio);

  if (error != 0)
    return rtems_bsd_error_to_status_and_errno(error);

  return td->td_retval[0];
}

ssize_t
rtems_bsd_sysgen_writev(rtems_libio_t      *iop,
                        const struct iovec *iov,
                        int                 iovcnt,
                        ssize_t             total)
{
  struct thread *td = curthread;
  struct uio auio;
  int error;

  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: writev: %d iocnt=%d len=%d\n",
           rtems_libio_iop_to_descriptor(iop), iovcnt, total);

  if (total > IOSIZE_MAX)
    return EINVAL;

  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: writev: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }

  auio.uio_iov = RTEMS_DECONST(struct iovec*, iov);
  auio.uio_iovcnt = iovcnt;
  auio.uio_resid = total;
  auio.uio_segflg = UIO_USERSPACE;

  error = kern_writev(td, rtems_bsd_libio_iop_to_descriptor(iop),
		      &auio);

  if (error != 0)
    return rtems_bsd_error_to_status_and_errno(error);

  return td->td_retval[0];
}

int
rtems_bsd_sysgen_ioctl(rtems_libio_t *iop, ioctl_command_t request, void *buffer)
{
  struct thread *td = curthread;
  u_long com = request & 0xffffffff;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: ioctl: %d req=%08x\n", rtems_libio_iop_to_descriptor(iop), com);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: ioctl: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  error = kern_ioctl(td, rtems_bsd_libio_iop_to_descriptor(iop),
                     com, buffer);
  return rtems_bsd_error_to_status_and_errno(error);
}

off_t
rtems_bsd_sysgen_lseek(rtems_libio_t *iop, off_t offset, int whence)
{
  struct thread *td = curthread;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: lseek: %d offset=%zu whence=%d\n",
           rtems_libio_iop_to_descriptor(iop), offset, whence);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: lseek: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  error = kern_lseek(td, rtems_bsd_libio_iop_to_descriptor(iop),
                     offset, whence);
  if (error != 0) {
    return rtems_bsd_error_to_status_and_errno(error);
  }
  return td->td_uretoff.tdu_off;
}

int
rtems_bsd_sysgen_vnstat(const rtems_filesystem_location_info_t *loc,
                        struct stat *buf)
{
  struct thread *td = curthread;
  struct vnode *vp = rtems_bsd_libio_loc_to_vnode(loc);
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: vnstat: %p\n", vp);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: vnstat: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  if (vp == NULL)
    error = EFAULT;
  else {
    VOP_LOCK(vp, LK_SHARED);
    error = vn_stat(vp, buf, td->td_ucred, NOCRED, td);
    VOP_UNLOCK(vp, 0);
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_fstat(const rtems_filesystem_location_info_t *loc,
                       struct stat *buf)
{
  struct thread *td = curthread;
  rtems_libio_t* iop = rtems_bsd_libio_loc_to_iop(loc);
  struct filedesc* fdp;
  struct file* fp = NULL;
  int fd;
  int error;
  if (iop == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: ffile: no iop\n");
    return rtems_bsd_error_to_status_and_errno(ENXIO);
  }
  fd = rtems_bsd_libio_iop_to_descriptor(iop);
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: fstat: %d\n", fd);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: fstat: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  fdp = td->td_proc->p_fd;
  FILEDESC_XLOCK(fdp);
  if (fd < fdp->fd_nfiles) {
    fp = fget_locked(fdp, fd);
  }
  FILEDESC_XUNLOCK(fdp);
  if (fp != NULL) {
    error = fo_stat(fp, buf, NULL, td);
  } else {
    error = EBADF;
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_ftruncate(rtems_libio_t *iop, off_t length)
{
  struct thread *td = curthread;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: ftruncate: len=%d\n", length);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: ftruncate: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  error = kern_ftruncate(td, rtems_bsd_libio_iop_to_descriptor(iop),
                         length);
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_fsync(rtems_libio_t *iop)
{
  struct thread *td = curthread;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: fsync\n");
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: fsync: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  error = kern_fsync(td, rtems_bsd_libio_iop_to_descriptor(iop),
                     true);
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_fdatasync(rtems_libio_t *iop)
{
  struct thread *td = curthread;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: fdatasync\n");
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: fdatasync: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  error = kern_fsync(td, rtems_bsd_libio_iop_to_descriptor(iop),
                     false);
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_fcntl(rtems_libio_t *iop, int cmd)
{
  struct thread *td = curthread;
  intptr_t arg;
  int error;
  if (RTEMS_BSD_SYSCALL_TRACE)
    printf("bsd: sys: fcntl: %d cmd=%d\n", rtems_libio_iop_to_descriptor(iop), cmd);
  if (td == NULL) {
    if (RTEMS_BSD_SYSCALL_TRACE)
      printf("bsd: sys: fcntl: no curthread\n");
    return rtems_bsd_error_to_status_and_errno(ENOMEM);
  }
  switch (cmd) {
  case F_SETFD:
    arg = rtems_libio_to_fcntl_flags(rtems_libio_iop_flags(iop)) & FD_CLOEXEC;
    break;
  case F_SETFL:
    arg = rtems_libio_to_fcntl_flags(rtems_libio_iop_flags(iop)) & FCNTLFLAGS;
    break;
  default:
    arg = -1;
    error = 0;
    break;
  }
  if (arg >= 0) {
    error = kern_fcntl(td, rtems_bsd_libio_iop_to_descriptor(iop),
		       cmd, arg);
    /* no return path with the RTEMS API for get calls */
  }
  return rtems_bsd_error_to_status_and_errno(error);
}

int
rtems_bsd_sysgen_poll(rtems_libio_t *iop, int events)
{
  printf("rtems_bsd_sysgen_poll called!\n");
  return rtems_bsd_error_to_status_and_errno(EOPNOTSUPP);
}

int
rtems_bsd_sysgen_kqfilter(rtems_libio_t *iop, struct knote *kn)
{
  printf("rtems_bsd_sysgen_kqfilter called!\n");
  return rtems_bsd_error_to_status_and_errno(EOPNOTSUPP);
}
